#
# Copyright (c) 2015 Midokura SARL, All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from charmhelpers.core import hookenv
from charmhelpers.core.services import base
from charmhelpers.core.services import helpers
from keystoneclient.v2_0 import client
from midonet_helpers import puppet
from midonet_helpers import relations as common_relations
from midonet_helpers import repositories
import relations

CREDS_PATH = '/etc/midonet_creds'


class ApiConfig(dict):
    def __init__(self, *args):
        self['config'] = hookenv.config()
        if not self['config']['api_ip']:
            self['config']['api_ip'] = hookenv.unit_get('public-address')


class TunnelHandling(base.ManagerCallback):
    def __call__(self, manager, service_name, event_name):
        service = manager.get_service(service_name)
        context = {}
        for ctx in service.get('required_data', []):
            context.update(ctx)
        self._add_to_tunnel_zone(context)

    def _host_addresses(self, conn):
        """Returns a map of physical address to host_id."""
        data = {}
        for host in conn.get_hosts():
            host_id = host.get_id()
            # Account for physical (nics) and bridges, reported as unknown
            physical_address_lists = [interface.get_addresses() for
                                      interface in host.get_interfaces() if
                                      interface.get_endpoint() in ('PHYSICAL',
                                                                   'UNKNOWN')]
            for physical_address_list in physical_address_lists:
                for address in physical_address_list:
                    data[address] = host_id
        return data

    def _default_tunnel_zone(self, conn, tunnel_type='gre', name='default_tz'):
        """Returns the default tunnel zone. Creates it if it did not exist."""
        tunnel_zones = [tz for tz in conn.get_tunnel_zones() if
                        tz.get_name() == name]
        if tunnel_zones:
            tz = tunnel_zones[0]
        else:
            tz = conn.add_tunnel_zone()
            tz.type(tunnel_type)
            tz.name(name)
            tz.create()
        return tz

    def _get_midonet_host_ips(self):
        """Generates an iterable of related ids to add to the tunnel zone."""
        for rid in hookenv.relation_ids('host'):
            for unit in hookenv.related_units(rid):
                yield hookenv.relation_get('private-address', unit=unit,
                                           rid=rid)

    def _get_keystone_token(self, context):
        keystone_client = client.Client(
            username=context['keystone'][0]['service_username'],
            password=context['keystone'][0]['service_password'],
            tenant_name=context['keystone'][0]['service_tenant'],
            auth_url='http://%s:%s/v2.0' %
            (context['keystone'][0]['auth_host'],
             context['keystone'][0]['auth_port']))
        return keystone_client.get_token(keystone_client.session)

    def _add_to_tunnel_zone(self, context):
        """Adds related host to the default tunnel zone."""
        host_addresses = list(self._get_midonet_host_ips())
        if not host_addresses:
            return
        try:
            from midonetclient import api
            api.auth_lib._token = self._get_keystone_token(context)
            conn = api.MidonetApi('http://%s:%s/midonet-api' %
                                  (context['config']['api_ip'],
                                   context['config']['api_port']),
                                  context['keystone'][0]['service_username'],
                                  context['keystone'][0]['service_password'],
                                  context['keystone'][0]['service_tenant'])
        except Exception:
            hookenv.log('Failed to connect to midonet-api. It may be stopped '
                        'will retry on the next event.')
            return

        try:
            address_map = self._host_addresses(conn)
        except Exception as ex:
            hookenv.log('Failed to get the address map. Err: %s' % ex)
            return

        for remote_address in host_addresses:
            host_id = address_map.get(remote_address)
            if host_id is None:
                hookenv.log('The host with address %s is not present in the '
                            'list of hosts in zookeeper. Midonet host agent '
                            'has probably not started there yet' %
                            remote_address)
                continue
            try:
                tz = self._default_tunnel_zone(conn)
                if any([host_id == host.get_host_id() for
                        host in tz.get_hosts()]):  # host already added
                    continue
                tz_host = tz.add_tunnel_zone_host()
                tz_host.ip_address(remote_address)
                tz_host.host_id(host_id)
                tz_host.create()
            except Exception as ex:
                hookenv.log('Failed to add host with address %s to the default'
                            ' tunnel zone. Err: %s' % (remote_address, ex))
                raise


class MidonetApiRelation(helpers.RelationContext):
    name = 'midonet-api'
    interface = 'midonet'
    port = 8080

    def __init__(self, *args, **kwargs):
        self.required_keys = []
        helpers.RelationContext.__init__(self, *args, **kwargs)

    def provide_data(self):
        return {
            'host': hookenv.unit_get('private-address'),
            'port': self.port,
        }


def manage():
    midonet_api_relation = MidonetApiRelation()
    cfg = ApiConfig()
    manager = base.ServiceManager([
        {
            'service': 'tomcat7',
            'ports': [midonet_api_relation.port],
            'provided_data': [midonet_api_relation],
            'required_data': [cfg,
                              relations.KeystoneRelation(),
                              common_relations.ZookeeperRelation(),
                              repositories.config(
                                  cfg['config'].get('midonet-release'),
                                  cfg['config'].get('mem-username'),
                                  cfg['config'].get('mem-password'))],
            'data_ready': [
                helpers.render_template(
                    source='common.yaml',
                    target='/var/lib/hiera/midonet_api/common.yaml'),
                puppet.InstallCallback('midonet::repository'),
                puppet.InstallCallback('midonet::midonet_api'),
                puppet.InstallCallback('midonet::midonet_cli'),
                helpers.render_template(
                    source='.midonetrc',
                    target='/root/.midonetrc'),
                helpers.render_template(
                    source='midonet-cli',
                    target='/usr/sbin/midonet-cli',
                    perms=0o777),  # midonet-cli workaround
                TunnelHandling(),
            ],
        },
    ])

    # ServiceManagers detect empty relations with no required keys as
    # incomplete. Thus, those relations that are entirely optional, should
    # only be added to 'required_data' if they bring data.
    if midonet_api_relation:
        manager.get_service('tomcat7').get('required_data', []).append(
            midonet_api_relation)
    manager.manage()
